#!/usr/bin/python3 -B
import argparse
import collections
import dataclasses
import json
import os
import pathlib
import subprocess
import sys
import textwrap
import time
from datetime import datetime

from debian.debian_support import DpkgArchTable

from debputy._deb_options_profiles import DebOptionsAndProfiles
from debputy.architecture_support import dpkg_architecture_table
from debputy.dh_migration import migrate_from_dh, AcceptableMigrationIssues
from debputy.packages import parse_source_debian_control
from debputy.version_substitutions import load_source_variables

# setup PYTHONPATH: add our installation directory.
sys.path.insert(0, pathlib.Path(__file__).parent)

from debputy.deb_packaging_support import setup_control_files, build_dbgsym_package_if_relevant
from debputy.highlevel_manifest import HighLevelManifest, ManifestPathInfoRuleType, ChangeDescription
from debputy.highlevel_manifest_parser import parse_manifest
from debputy.intermediate_manifest import TarMember
from debputy.substitution import Substitution, SubstitutionImpl, NULL_SUBSTITUTION
from debputy.util import _error, compute_output_filename

try:
    from typing import Union, NoReturn, List, Tuple, Dict, Iterable, Optional, Set, Callable, Any

    ManifestKey = Union[str, Tuple[str, str]]
except ImportError:
    pass


def parse_args():
    description = textwrap.dedent(f'''\
    The `debputy` program is a manifest-based Debian packaging tool. 
    
    It is used as a part of compiling a source package and transforming it into one or
    more binary (.deb) packages.
    ''')

    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        allow_abbrev=False,
    )
    parser.add_argument('--package-manifest', dest='package_manifest',
                        action='store', default=None,
                        help='Manifest of file metadata that is not visible in the file-system')
    parser.add_argument('--substitution', dest='substitution',
                        action='store_true', default=None,
                        help='*Some* commands can work without doing path substitution.  With this option, substitution'
                             ' is always done.  It is a no-op for commands that requires substitution'
                        )
    parser.add_argument('--no-substitution', dest='substitution',
                        action='store_false', default=None,
                        help='*Some* commands can work without doing path substitution.  With this option, substitution'
                             ' is not done.  However, this option *cannot* be used with all commands.'
                        )

    parser.add_argument('-p', '--package', dest='packages',
                        action='append', type=str, default=[],
                        help='The package(s) to act on.  Affects default permission normalization rules'
                        )

    subparsers = parser.add_subparsers(dest='command', required=True)

    generate_deb_package = subparsers.add_parser('generate-debs-from-binary-staging-directories',
                                                 help="Generate .deb/.udebs packages from debian/<pkg>")
    generate_deb_package.add_argument('output', metavar="output",
                                      help='Where to place the resulting packages.  Should be a directory'
                                      )
    # Added for "help only" - you cannot trigger this option in practice
    generate_deb_package.add_argument('--', metavar='UPSTREAM_ARGS', action='extend', nargs='+', dest='unused')

    migrate_from_debhelper = subparsers.add_parser('migrate-from-dh',
                                                   help='Generate/update manifest from a "dh $@" using package')
    migrate_from_debhelper.add_argument('--acceptable-migration-issues', dest='acceptable_migration_issues',
                                        action='append', type=str, default=[],
                                        help='Continue the migration even if this/these issues are detected.'
                                             ' Can be set to ALL (in all upper-case) to accept all issues')

    migrate_from_debhelper.add_argument('--no-act', dest='destructive',
                                        action='store_false', default=True,
                                        help='Do not perform changes.  Existing manifest will not be overridden')

    show_changes_parser = subparsers.add_parser('show-manifest-changes', help="Show what the manifest will do")
    show_changes_parser.add_argument('--show-implicit-changes', action='store_true', default=False,
                                     dest='show_implicit_changes',
                                     help="Show implicit changes (usually directories implied by other actions)")
    show_changes_parser.add_argument('--show-builtin-rules', action='store_true', default=False,
                                     dest='show_builtin_rules',
                                     help="Show builtin rules (e.g., default permission normalization)")

    generator_parser = subparsers.add_parser('debug-generate-intermediate-manifest',
                                             help="[Debugging] Generate intermediate manifest for deb_packer")

    generator_parser.add_argument('package_root_dir', metavar="PACKAGE_ROOT_DIR",
                                  help='Root directory of the package. Must contain a DEBIAN directory'
                                  )
    generator_parser.add_argument('intermediate_manifest_output', nargs='?', metavar="PATH",
                                  help='Path where the intermediate manifest should be placed.'
                                  )

    argv = sys.argv
    try:
        i = argv.index('--')
        upstream_args = argv[i + 1:]
        argv = argv[:i]
    except (IndexError, ValueError):
        upstream_args = []
    parsed_args = parser.parse_args(argv[1:])
    setattr(parsed_args, 'upstream_args', upstream_args)
    setattr(parsed_args, 'packages', frozenset(parsed_args.packages))

    return parsed_args


def _serialize_intermediate_manifest(members: Iterable[TarMember]) -> str:
    serial_format = [m.to_manifest() for m in members]
    return json.dumps(serial_format)


def output_intermediate_manifest(manifest_output_file: str, members: Iterable[TarMember]) -> None:
    with open(manifest_output_file, 'w') as fd:
        serial_format = [m.to_manifest() for m in members]
        json.dump(serial_format, fd)


def _generate_intermediate_manifest(parsed_args, manifest: HighLevelManifest, mtime: int) -> None:
    output_path = parsed_args.intermediate_manifest_output
    active_packages = list(manifest.active_packages)
    if len(active_packages) != 1:
        _error("This command requires exactly one package (-p/--package")
    root_dir = parsed_args.package_root_dir
    package = active_packages[0].name
    _, intermediate_manifest = manifest.apply_to_binary_staging_directory(package, root_dir, mtime)
    output_intermediate_manifest(output_path, intermediate_manifest)


def _print_manifest_change_group(grouped: Dict[ManifestPathInfoRuleType, List[ChangeDescription]],
                                 rule_type: ManifestPathInfoRuleType,
                                 header: str,
                                 footer: str = ''
                                 ):
    changes = grouped.get(rule_type)
    if changes:
        print(header)
        for change_description in changes:
            print(change_description.pretty_format())
        print()
        if footer:
            print(footer)
            print()


def _show_manifest_changes(parsed_args, manifest: HighLevelManifest, mtime: int) -> None:
    for dctrl_bin in manifest.active_packages:
        package = dctrl_bin.name
        package_state = manifest.package_state_for(package)
        declared_transformations = [t for t in package_state.transformations if not t.is_builtin_rule]

        if declared_transformations:
            print("Transformations:")
            for transformation in declared_transformations:
                print(f' * {transformation.describe_transformation()}')
            print()
        grouped = collections.defaultdict(list)
        for change_description in sorted(manifest.change_descriptions(package),
                                         key=lambda x: x.match_rule.sort_key(),
                                         reverse=True,
                                         ):
            path_info = change_description.path_info
            if path_info.is_builtin_definition and not parsed_args.show_builtin_rules:
                continue
            if path_info.is_implicit_definition and not parsed_args.show_implicit_changes:
                continue
            grouped[change_description.path_info.rule_type].append(change_description)

        _print_manifest_change_group(grouped, ManifestPathInfoRuleType.CREATE_DIRECTORY, 'Directory creation:')
        _print_manifest_change_group(grouped, ManifestPathInfoRuleType.CREATE_SYMLINK, 'Symlink creation:')

        print("Metadata changes (default settings):")
        print(" * Owner: root (uid: 0)")
        print(" * Group: root (gid: 0)")
        print(f" * mtime (clamp/upper bound): {mtime} ({datetime.utcfromtimestamp(mtime)}Z)")
        print()
        _print_manifest_change_group(grouped, ManifestPathInfoRuleType.ENSURE_METADATA, 'Metadata changes (path rules):')

        if package_state.maintscript_snippets:
            print()
            print("Maintscript snippets triggered by the manifest:")
            for script, snippets in package_state.maintscript_snippets.items():
                print(f" * {script}:")
                for snippet in snippets.all_snippets():
                    print(f"     - triggered by {snippet.definition_source}")


def _generate_debs_from_binary_staging_directories(parsed_args, manifest: HighLevelManifest, mtime: int) -> None:
    output_path = parsed_args.output
    upstream_args = parsed_args.upstream_args

    for dctrl_bin in manifest.active_packages:
        package = dctrl_bin.name
        root_dir = os.path.join('debian', package)
        fs_root, intermediate_manifest = manifest.apply_to_binary_staging_directory(package, root_dir, mtime)

        assert dctrl_bin.should_be_acted_on

        setup_control_files(dctrl_bin, root_dir, fs_root, manifest)

        build_dbgsym_package_if_relevant(dctrl_bin, upstream_args, output_path)

        deb_packer = os.path.join(pathlib.Path(__file__).parent, 'deb_packer.py')
        deb_cmd = [deb_packer, '--intermediate-package-manifest', '-', '--source-date-epoch', str(mtime)]
        output = output_path
        if dctrl_bin.is_udeb:
            deb_cmd.extend(['-z6', '-Zxz', '-Sextreme'])
            output = os.path.join(output_path, compute_output_filename(root_dir, True))
        deb_cmd.extend(upstream_args)
        deb_cmd.extend(['--build', root_dir, output])

        print(f"Executing for {package}: {deb_cmd}")
        proc = subprocess.Popen(deb_cmd, stdin=subprocess.PIPE)
        proc.communicate(_serialize_intermediate_manifest(intermediate_manifest).encode('utf-8'))
        if proc.returncode != 0:
            _error(f"{deb_packer} exited with a non-zero exit code!")


def _migrate_from_dh(parsed_args, manifest: HighLevelManifest, _mtime: int) -> None:
    acceptable_migration_issues = AcceptableMigrationIssues(
        frozenset(i for x in parsed_args.acceptable_migration_issues for i in x.split(','))
    )
    migrate_from_dh(manifest, acceptable_migration_issues, parsed_args.destructive)


@dataclasses.dataclass
class Command:

    handler: Callable[[Any, HighLevelManifest, int], None]
    require_substitution: bool = True

    def create_substitution(self, parsed_args, extra_substitutions: Dict[str, str]) -> Substitution:
        requested_subst = parsed_args.substitution
        if requested_subst is False and self.require_substitution:
            _error(f"--no-substitution cannot be used with {parsed_args.command}")
        if self.require_substitution or requested_subst is not False:
            if not os.path.isfile('debian/changelog'):
                _error('Expected to be run from a directory containing debian/changelog')
            return SubstitutionImpl(extra_substitutions=extra_substitutions)
        return NULL_SUBSTITUTION


COMMANDS = {
    'debug-generate-intermediate-manifest': Command(handler=_generate_intermediate_manifest),
    'generate-debs-from-binary-staging-directories': Command(handler=_generate_debs_from_binary_staging_directories),
    'show-manifest-changes': Command(handler=_show_manifest_changes, require_substitution=False),
    'migrate-from-dh': Command(handler=_migrate_from_dh)
}


def main():
    parsed_args = parse_args()

    command = COMMANDS[parsed_args.command]

    packages = parsed_args.packages
    v = load_source_variables('debian/changelog')
    substitution = command.create_substitution(parsed_args, v)
    if 'SOURCE_DATE_EPOCH' not in os.environ:
        os.environ['SOURCE_DATE_EPOCH'] = v['SOURCE_DATE_EPOCH']
    mtime = int(os.environ['SOURCE_DATE_EPOCH'])
    build_env = DebOptionsAndProfiles.instance()
    dpkg_architecture_variables = dpkg_architecture_table()
    dpkg_arch_query_table = DpkgArchTable.load_arch_table()

    _, binary_packages = parse_source_debian_control(
        packages,  # -p/--package
        set(),  # -N/--no-package
        False,  # -i
        False,  # -a
        dpkg_architecture_variables=dpkg_architecture_variables,
        dpkg_arch_query_table=dpkg_arch_query_table,
        build_env=build_env,
    )
    assert packages <= binary_packages.keys()

    manifest = parse_manifest(parsed_args.package_manifest,
                              binary_packages,
                              substitution,
                              dpkg_architecture_variables,
                              dpkg_arch_query_table,
                              build_env,
                              )
    command.handler(parsed_args, manifest, mtime)


if __name__ == '__main__':
    main()
