import dataclasses
import json
import os
import subprocess
from typing import Sequence, Iterable, Optional, FrozenSet, List, Tuple

from debputy.debhelper_emulation import dhe_pkgfile, dhe_filedoublearray, DHConfigFileLine
from debputy.highlevel_manifest import HighLevelManifest, MutableYAMLSymlink, MutableYAMLConffileManagementItem, \
    MutableCondition
from debputy.installdeb_emulation import _validate_rm_mv_conffile
from debputy.packages import BinaryPackage
from debputy.substitution import NULL_SUBSTITUTION
from debputy.util import _error


class AcceptableMigrationIssues:
    def __init__(self, values: FrozenSet[str]):
        self._values = values

    def __contains__(self, item: str) -> bool:
        return item in self._values or 'ALL' in self._values


class UnsupportedFeature(RuntimeError):

    @property
    def message(self) -> str:
        return self.args[0]

    @property
    def issue_keys(self) -> Optional[Sequence[str]]:
        if len(self.args) < 2:
            return None
        return self.args[1]


class ConflictingChange(RuntimeError):
    @property
    def message(self) -> str:
        return self.args[0]


@dataclasses.dataclass
class FeatureMigration:
    tagline: str
    successful_changes: int
    already_present: int
    warnings: Sequence[str] = tuple()
    remove_paths_on_success: Sequence[str] = tuple()

    @property
    def anything_to_do(self) -> bool:
        return bool(self.total_changes_involved)

    @property
    def total_changes_involved(self) -> int:
        return self.successful_changes + len(self.warnings) + len(self.remove_paths_on_success)


def _extract_condition_from_dh_config_line(dh_config_line: DHConfigFileLine,
                                           acceptable_migration_issues: AcceptableMigrationIssues,
                                           warnings: List[str]):
    if dh_config_line.build_profile_filter and dh_config_line.arch_filter:
        if 'dh-exec-multiple-filters' in acceptable_migration_issues:
            warnings.append('TODO: MANUAL MIGRATION of conditions from executable dh config'
                            f' "{dh_config_line.config_file}" line {dh_config_line.line_no} is required')
            return None
        raise UnsupportedFeature('The dh-exec emulation cannot support multiple filters at the moment'
                                 f' (found: {dh_config_line.config_file} at line {dh_config_line.line_no}',
                                 ['dh-exec-multiple-filters'],
                                 )
    if dh_config_line.build_profile_filter:
        return MutableCondition.build_profiles_matches(dh_config_line.build_profile_filter)
    if dh_config_line.arch_filter:
        return MutableCondition.arch_matches(dh_config_line.arch_filter)
    return None


def _dh_config_file(dctrl_bin: BinaryPackage,
                    basename,
                    helper_name,
                    acceptable_migration_issues: AcceptableMigrationIssues,
                    warnings: List[str],
                    paths_to_remove: List[str],
                    support_executable_files=False,
                    ) -> Tuple[Optional[str], Optional[Iterable[DHConfigFileLine]]]:
    dh_config_file = dhe_pkgfile(dctrl_bin, basename)
    if dh_config_file is None:
        return None, None
    if os.access(dh_config_file, os.X_OK) and not support_executable_files:
        primary_key = f'executable-{helper_name}-config'
        if (primary_key in acceptable_migration_issues
                or 'any-executable-dh-configs' in acceptable_migration_issues):
            warnings.append(f'TODO: MANUAL MIGRATION of executable dh config "{dh_config_file}" is required.')
            return None, None
        raise UnsupportedFeature(f'Executable configuration files not supported (found: {dh_config_file}).',
                                 [primary_key, 'any-executable-dh-configs'],
                                 )

    paths_to_remove.append(dh_config_file)
    content = dhe_filedoublearray(dh_config_file, NULL_SUBSTITUTION)
    return dh_config_file, content


def migrate_maintscript(manifest: HighLevelManifest,
                        acceptable_migration_issues: AcceptableMigrationIssues,
                        ) -> Iterable[FeatureMigration]:
    mutable_manifest = manifest.mutable_manifest
    paths_to_remove = []
    warnings = []
    manifest_changes = 0
    already_present = 0
    for dctrl_bin in manifest.all_packages:
        mainscript_file, content = _dh_config_file(dctrl_bin,
                                                   'maintscript',
                                                   'dh_installdeb',
                                                   acceptable_migration_issues,
                                                   warnings,
                                                   paths_to_remove,
                                                   )

        if mainscript_file is None:
            continue
        assert content is not None

        package_definition = mutable_manifest.package(dctrl_bin.name)
        conffiles = {it.obsolete_conffile: it for it in package_definition.conffile_management_items()}
        seen_conffiles = set()

        for dhe_line in content:
            cmd = dhe_line.tokens[0]
            if cmd not in {'rm_conffile', 'mv_conffile'}:
                raise UnsupportedFeature(f'The dh_installdeb file {mainscript_file} contains the (currently)'
                                         f' unsupported command "{cmd}" on line {dhe_line.line_no}'
                                         f' (line: "{dhe_line.original_line}"')

            try:
                _, obsolete_conffile, new_conffile, prior_to_version, owning_package = \
                    _validate_rm_mv_conffile(dctrl_bin.name, dhe_line)
            except ValueError as e:
                _error(f"Validation error in {mainscript_file} on line {dhe_line.line_no}. The error was: {e.args[0]}.")

            if obsolete_conffile in seen_conffiles:
                raise ConflictingChange(f'The {mainscript_file} file defines actions for "{obsolete_conffile}" twice!'
                                        f' Please ensure that it is defined at most once in that file.')
            seen_conffiles.add(obsolete_conffile)

            if cmd == 'rm_conffile':
                item = MutableYAMLConffileManagementItem.rm_conffile(
                    obsolete_conffile,
                    prior_to_version,
                    owning_package,
                )
            else:
                assert cmd == 'mv_conffile'
                item = MutableYAMLConffileManagementItem.mv_conffile(
                    obsolete_conffile,
                    new_conffile,
                    prior_to_version,
                    owning_package,
                )

            existing_def = conffiles.get(item.obsolete_conffile)
            if existing_def is not None:
                if not (
                        item.command == existing_def.command
                        and item.new_conffile == existing_def.new_conffile
                        and item.prior_to_version == existing_def.prior_to_version
                        and item.owning_package == existing_def.owning_package
                ):
                    raise ConflictingChange(f'The maintscript defines the action {item.command} for'
                                            f' "{obsolete_conffile}" in {mainscript_file}, but there is another'
                                            f' conffile management definition for same path defined already (in the'
                                            f' existing manifest or an migration e.g., inside {mainscript_file})')
                already_present += 1
                continue

            package_definition.add_conffile_management(item)
            manifest_changes += 1

    yield FeatureMigration('dh_installdeb files',
                           warnings=warnings,
                           remove_paths_on_success=paths_to_remove,
                           successful_changes=manifest_changes,
                           already_present=already_present,
                           )


def migrate_tmpfile(manifest: HighLevelManifest,
                    acceptable_migration_issues: AcceptableMigrationIssues,
                    ) -> Iterable[FeatureMigration]:
    warnings = []
    for dctrl_bin in manifest.all_packages:
        dh_config_file = dhe_pkgfile(dctrl_bin, 'tmpfile')
        if dh_config_file is not None:
            # TODO: Support migrating the file to its new location.
            if 'legacy-dh_installtmpfiles-config-name' not in acceptable_migration_issues:
                raise UnsupportedFeature(f'The dh_installtmpfiles file {dh_config_file} is using the "legacy" name. '
                                         ' Please use the ".tmpfiles" (note the "s" at the end) name instead.  This is'
                                         ' also supported by debhelper assuming you are using dh_installtmpfiles'
                                         ' (default in compat 13 or newer).')
            warnings.append(f'TODO: MANUAL MIGRATION of legacy dh_installtmpfiles config "{dh_config_file}" is'
                            ' required.')
    yield FeatureMigration('dh_installtmpfiles config files',
                           warnings=warnings,
                           remove_paths_on_success=tuple(),
                           successful_changes=0,
                           already_present=0,
                           )


def migrate_links_files(manifest: HighLevelManifest,
                        acceptable_migration_issues: AcceptableMigrationIssues,
                        ) -> Iterable[FeatureMigration]:
    mutable_manifest = manifest.mutable_manifest
    paths_to_remove = []
    warnings = []
    manifest_changes = 0
    already_present = 0
    for dctrl_bin in manifest.all_packages:
        links_file, content = _dh_config_file(dctrl_bin,
                                              'links',
                                              'dh_link',
                                              acceptable_migration_issues,
                                              warnings,
                                              paths_to_remove,
                                              support_executable_files=True,
                                              )

        if links_file is None:
            continue
        assert content is not None

        package_definition = mutable_manifest.package(dctrl_bin.name)
        defined_symlink = {symlink.symlink_path: symlink.symlink_target for symlink in package_definition.symlinks()}

        seen_symlinks = set()

        for dhe_line in content:
            if len(dhe_line.tokens) != 2:
                raise UnsupportedFeature(f'The dh_link file {links_file} did not have exactly two paths on line'
                                         f' {dhe_line.line_no} (line: "{dhe_line.original_line}"')
            target, source = dhe_line.tokens
            if source in seen_symlinks:
                # According to #934499, this has happened in the wild already
                raise ConflictingChange(f'The {links_file} file defines the link path {source} twice! Please ensure'
                                        ' that it is defined at most once in that file')
            seen_symlinks.add(source)
            # Symlinks in .links are always considered absolute, but you were not required to have a leading slash.
            # However, in the debputy manifest, you can have relative links, so we should ensure it is explicitly
            # absolute.
            if not target.startswith('/'):
                target = '/' + target
            existing_target = defined_symlink.get(source)
            if existing_target is not None:
                if existing_target != target:
                    raise ConflictingChange(f'The symlink "{source}" points to "{target}" in {links_file}, but there is'
                                            f' another symlink with same path pointing to "{existing_target}" defined'
                                            ' already (in the existing manifest or an migration e.g., inside'
                                            f' {links_file})')
                already_present += 1
                continue
            condition = _extract_condition_from_dh_config_line(dhe_line,
                                                               acceptable_migration_issues,
                                                               warnings,
                                                               )
            package_definition.add_symlink(MutableYAMLSymlink.new_symlink(source,
                                                                          target,
                                                                          condition,
                                                                          ))
            manifest_changes += 1
    yield FeatureMigration('dh_link files',
                           warnings=warnings,
                           remove_paths_on_success=paths_to_remove,
                           successful_changes=manifest_changes,
                           already_present=already_present,
                           )


def _print_migration_summary(migrations: List[FeatureMigration],
                             permit_destructive_changes: bool,
                             compat: int,
                             min_compat_level: int,
                             ) -> None:
    warning_count = 0
    print("MIGRATION SUMMERY")
    print("=================")
    print()

    for migration in migrations:
        print()
        if not migration.anything_to_do:
            print(f'Migration: {migration.tagline} - Nothing to do and no issues detected!')
            continue
        underline = '-' * len(migration.tagline)
        print(f'Summary for migration: {migration.tagline}')
        print(f'-----------------------{underline}')
        print()
        if migration.warnings:
            print(" /!\\ ATTENTION /!\\")
            warning_count += len(migration.warnings)
            for warning in migration.warnings:
                print(f"    * {warning}")
        else:
            print(f'  No warnings detected / manual conversions needed')
        print()
        if permit_destructive_changes:
            print(f' * Manifest changes applied: {migration.successful_changes}')
            print(f' * Settings already present: {migration.already_present}')
            print(f' * Files removed (merged): {migration.remove_paths_on_success}')
        else:
            print(f' * Manifest changes to be applied: {migration.successful_changes}')
            print(f' * Settings already present: {migration.already_present}')
            print(f' * Files to be removed (merged): {migration.remove_paths_on_success}')

    if compat < min_compat_level:
        warning_count += 1
        print()
        print(f"WARNING: The migration tool assumes debhelper compat {min_compat_level}+ semantics, but this package"
              f" is using compat {min_compat_level}.  Consider upgrading the package to compat {min_compat_level}"
              " first.")

    if warning_count:
        print()
        print(f'/!\\ Total number of warnings or manual migrations required: {warning_count}')


def _dh_compat_level() -> Optional[int]:
    try:
        res = subprocess.check_output(['dh_assistant', 'active-compat-level'], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        compat = None
    else:
        try:
            compat = json.loads(res)["declared-compat-level"]
        except RuntimeError:
            compat = None
        else:
            if not isinstance(compat, int):
                compat = None
    return compat


def migrate_from_dh(manifest: HighLevelManifest,
                    acceptable_migration_issues: AcceptableMigrationIssues,
                    permit_destructive_changes: bool) -> None:
    migrations = []
    compat = _dh_compat_level()
    # Compat 14 due to dh_installpam installation directory
    # Compat 13 due to dh_installtmpfiles (see TODO in migrate_tmpfile)
    min_compat = 14
    if compat is None:
        _error(f'Cannot detect declared compat level (try running "dh_assistant active-compat-level")')
        return

    if compat < min_compat and 'min-compat-level' not in acceptable_migration_issues:
        # The migration summary special-cases the compat mismatch and warns for us.
        _error(f'The migration tool assumes debhelper compat {min_compat} or later but the package is only on'
               f' compat {compat}.  This may lead to incorrect result.'
               f"\n\nUse --acceptable-migration-issues=min-compat-level to convert this into a warning and"
               f" try again, if you want to continue regardless.")

    try:
        migrations.extend(migrate_links_files(manifest, acceptable_migration_issues))
        migrations.extend(migrate_maintscript(manifest, acceptable_migration_issues))
        migrations.extend(migrate_tmpfile(manifest, acceptable_migration_issues))
    except UnsupportedFeature as e:
        msg = f'Unable to migrate automatically due to missing features in debputy. The feature is:' \
              f'\n\n  * {e.message}'
        keys = e.issue_keys
        if keys:
            primary_key = keys[0]
            alt_keys = ''
            if len(keys) > 1:
                alt_keys = f' Alternatively you can also use one of: {", ".join(keys[1:])}.  Please note that some' \
                           ' of these may cover more cases.'
            msg += (f"\n\nUse --acceptable-migration-issues={primary_key} to convert this into a warning and try again."
                    " However, you should only do that if you believe you can replace the functionality manually"
                    f" or the usage is obsolete / can be removed. {alt_keys}"
                    )
        _error(msg)
    except ConflictingChange as e:
        _error("The migration tool detected a conflict data being migrated and data already migrated / in the existing"
               "manifest."
               f"\n\n   * {e.message}"
               '\n\nPlease review the situation and resolve the conflict manually.')

    _print_migration_summary(migrations, permit_destructive_changes, compat, min_compat)
    manifest_change_count = sum((m.successful_changes for m in migrations), 0)

    if not manifest_change_count:
        print()
        print(f"debputy was not able to find any (supported) migrations that it could perform for you.")
        return

    with open(manifest.manifest_path + ".new", 'w') as fd:
        manifest.mutable_manifest.write_to(fd)

    if permit_destructive_changes:
        if os.path.isfile(manifest.manifest_path):
            os.rename(manifest.manifest_path, manifest.manifest_path + '.orig')
        os.rename(manifest.manifest_path + '.new', manifest.manifest_path)
        print()
        print(f"Updated manifest {manifest.manifest_path}")
    else:
        print()
        print(f'Created draft manifest "{manifest.manifest_path}.new" (rename to "{manifest.manifest_path}"'
              ' to activate it)')

    if permit_destructive_changes:
        for path in (p for m in migrations for p in m.remove_paths_on_success):
            os.unlink(path)
    else:
        c = sum((len(m.remove_paths_on_success) for m in migrations), 0)
        print(f"(Kept {c} file/files in place as requested)")
