from enum import Enum
from typing import Callable


def _named(prefix=None, suffix=None) -> Callable[[str], str]:

    def _impl(name: str) -> str:
        dest_name = name
        if prefix is not None and not dest_name.startswith(prefix):
            dest_name = prefix + dest_name
        if suffix is not None and not dest_name.endswith(suffix):
            dest_name += suffix
        return dest_name

    return _impl


def _replace_dot_with_underscore(x: str) -> str:
    return x.replace('.', '_')


class PackagerProvidedFileClass(Enum):
    SYSTEMD_TMPFILES_FILE = ('tmpfiles', './usr/lib/tmpfiles.d', 0o0644, _named(suffix='.conf'))
    PAM_CONFIG_FILE = ('pam', './usr/lib/pam.d', 0o0644, _named())
    PPP_IP_UP_FILE = ('ppp.ip-up', './etc/ppp/ip-up.d', 0o0755, _named())
    PPP_IP_DOWN_FILE = ('ppp.ip-down', './etc/ppp/ip-down.d', 0o0755, _named())
    LOGROTATE_CONFIG_FILE = ('logrotate', './etc/logrotate.d', 0o0644, _named())

    LOGCHECK_CRACKING_FILE = ('logcheck.cracking', './etc/logcheck/cracking.d', 0o644, _replace_dot_with_underscore)
    LOGCHECK_VIOLATIONS_FILE = ('logcheck.violations', './etc/logcheck/violations.d', 0o644,
                                _replace_dot_with_underscore)
    LOGCHECK_VIOLATIONS_IGNORE_FILE = ('logcheck.violations.ignore', './etc/logcheck/violations.ignore.d', 0o644,
                                       _replace_dot_with_underscore)
    LOGCHECK_IGNORE_WORKSTATION_FILE = ('logcheck.ignore.workstation', './etc/logcheck/ignore.d.workstation', 0o644,
                                        _replace_dot_with_underscore)
    LOGCHECK_IGNORE_SERVER_FILE = ('logcheck.ignore.server', './etc/logcheck/ignore.d.server', 0o644,
                                   _replace_dot_with_underscore)
    LOGCHECK_IGNORE_PARANOID_FILE = ('logcheck.ignore.paranoid', './etc/logcheck/ignore.d.paranoid', 0o644,
                                     _replace_dot_with_underscore)

    MIME_CONFIG_FILE = ('mime', './usr/lib/mime/packages/', 0o0644, _named())
    SHARED_MIME_INFO_CONFIG_FILE = ('sharedmimeinfo', './usr/share/mime/packages', 0o0644, _named(suffix='.xml'))

    IF_PRE_UP_CONFIG_FILE = ('if-pre-up', './etc/network/if-pre-up.d/', 0o0755, _named())
    IF_UP_CONFIG_FILE = ('if-up', './etc/network/if-up.d/', 0o0755, _named())
    IF_DOWN_CONFIG_FILE = ('if-down', './etc/network/if-down.d/', 0o0755, _named())
    IF_POST_DOWN_CONFIG_FILE = ('if-post-down', './etc/network/if-post-down.d/', 0o0755, _named())

    CRON_HOURLY_FILE = ('cron.hourly', './etc/cron.hourly', 0o0755, _named())
    CRON_DAILY_FILE = ('cron.daily', './etc/cron.daily', 0o0755, _named())
    CRON_WEEKLY_FILE = ('cron.weekly', './etc/cron.weekly', 0o0755, _named())
    CRON_MONTHLY_FILE = ('cron.monthly', './etc/cron.monthly', 0o0755, _named())
    CRON_YEARLY_FILE = ('cron.yearly', './etc/cron.yearly', 0o0755, _named())
    # cron.d uses 0644 unlike the others
    CRON_D_CONFIG_FILE = ('cron.d', './etc/cron.d', 0o0644, _named())

    INITRAMFS_HOOK_FILE = ('initramfs-hook', './usr/share/initramfs-tools/hooks', 0o0755, _named())

    MODPROBE_FILE = ('modprobe', './etc/modprobe.d', 0o0644, _named(suffix='.conf'))

    @property
    def default_lookup_basename(self) -> str:
        return self.value[0]

    @property
    def install_into_directory(self) -> str:
        return self.value[1]

    @property
    def default_mode(self) -> int:
        return self.value[2]

    @property
    def compute_dest_name(self) -> Callable[[str], str]:
        return self.value[3]


PACKAGER_PROVIDED_FILE_DEFINITIONS = {
    x.default_lookup_basename: x for x in PackagerProvidedFileClass
}

del _named
